; Prosty program boot loader-a dla interfejsu RS-232
; Rafa Baranowski (C) 2004

.INCLUDE "m16def.inc"	; SEGMENT DEKLARACJI

.EQU OSC_FREQ	= 16000000	; czstotliwo pracy w Hz
.EQU RS_BAUD	= 38400	;   prdko transmisji dla RS-232 [bps]
					; stae wskanikw LED
.EQU K_LEDY		= DDRD	;   rejestr kierunku portu
.EQU O_LEDY		= PORTD	;   rejestr wyjciowy portu
.EQU LED_ACT	= 6		;   nr linii dla wskanika aktywnoci boot loader-a
.EQU LED_ERR	= 7		;   nr linii dla wskanika bdu boot loader-a
					; stae klawiatury
.EQU K_KLAW		= DDRC	;   rejestr kierunku portu
.EQU O_KLAW		= PORTC	;   rejestr wyjciowy portu
.EQU I_KLAW		= PINC	;   rejestr wejciowy portu
.EQU KL_1		= 0		;   nr linii dla klawisza 1 (weryfikacja)
.EQU KL_4		= 1		;   nr linii dla klawisza 4 (programowanie)

.EQU ILE_STRON	= LARGEBOOTSTART/PAGESIZE	; definicja liczby stron sekcji aplikacji

					; nazwy symboliczne rejestrw roboczych
.DEF STR_ZH		= R25		;   MSB wskanika zapisywanej strony
.DEF STR_ZL		= R24		;   LSB wskanika zapisywanej strony 


.DSEG			; SEGMENT DANYCH
.ORG	0x0060			; pocztek pamici SRAM
Bufor:	.BYTE	2*PAGESIZE	; pomocniczy bufor strony w SRAM


.MACRO ldi_wsk			; makro adujce wskanik 16-bitowy sta 16-bitow
					; parametry:
					;   0. modyfikowany wskanik (X, Y lub Z)
					;   1. 16-bitowa bitowa warto bezporednia
	ldi	@0H, high(@1)
	ldi	@0L, low(@1)
.ENDM


.CSEG			; SEGMENT PROGRAMU
.ORG	0				; POCZTEK SEKCJI APLIKACJI
	jmp	0			; ptla nieskoczona (program tymczasowy)


.ORG	LARGEBOOTSTART		; POCZTEK SEKCJI BOOT LOADER-A (BLS)
					; (wektor inicjalizacji, poniewa BOOTRST=0)
Boot_loader:
	cli				; blokowanie przerwa

	ldi_wsk X, RAMEND
	out	SPH, XH
	out	SPL, XL		; inicjalizacja wskanika stosu

	ldi	R16, 0
	out	K_KLAW, R16
	ldi	R16, (1<<KL_1)|(1<<KL_4)
	out	O_KLAW, R16		; inicjalizacja maej klawiatury
					;   (wykorzystywane klawisze: "1" i "4")
	nop
	nop				; oczekuj na stabilizacj stanu na liniach klawiatury
	
	sbic	I_KLAW, KL_1	; jeli klawisz "1" nie jest przytrzymywany, to..
	jmp	0			;   przejd do wykonywania programu aplikacji


	ldi	R16, high(OSC_FREQ/(16*RS_BAUD)-1)
	out	UBRRH, R16
	ldi	R16, low(OSC_FREQ/(16*RS_BAUD)-1)
	out	UBRRL, R16		; ustal prdko dziaania moduu USART

	ldi	R16, (1<<RXEN)|(1<<TXEN)
	out	UCSRB, R16		; uaktywnij nadajnik i odbiornik USART
	ldi	R16, (1<<URSEL)|(1<<UCSZ1)|(1<<UCSZ0)|(1<<UPM1)|(1<<UPM0)
	out	UCSRC, R16		; wcz USART w trybie asynchronicznym, z 8-bitow
					;   dugoci sowa, 1 bitem stopu
					;   i bitem parzystoci typu odd

	sbic	I_KLAW, KL_4	; jeli klawisz "4" nie jest przytrzymywany, to..
	rjmp	Weryfikacja		;   przejd do programu weryfikujcego

					; WYMAZYWANIE ZAWARTOCI SEKCJI APLIKACJI
	ldi_wsk X, ILE_STRON	; inicjalizuj licznik stron sekcji aplikacji
	ldi_wsk Y, PAGESIZE<<1	; zapisz w Y liczb bajtw tworzcych jedn stron
	ldi_wsk Z, 0		; inicjalizuj wskanik zapisu pamici flash
Wymazywanie:
	ldi	R16, (1<<PGERS)|(1<<SPMEN)
	rcall	Wykonaj_spm		; wyma kolejn stron pamici flash

	add	ZL, YL		; zwiksz wskanik Z tak, by wskazywa on na kolejn
	adc	ZH, YH		;   stron pamici flash
	sbiw	XL, 1			; dekrementuj licznik pozostaych do wymazania stron
	brne	Wymazywanie		; zaptl iteracyjne (a do wymazania caej
					;   sekcji aplikacji)

	sbi	K_LEDY, LED_ACT
	sbi	O_LEDY, LED_ACT	; zapal wskanik aktywnoci boot loader-a

					; PROGRAMOWANIE ZAWARTOCI SEKCJI APLIKACJI
	ldi_wsk X, ILE_STRON	; inicjalizuj licznik stron sekcji aplikacji
	ldi_wsk STR_Z, 0		; inicjalizuj pomocniczy wskanik stron w pamici flash

Petla_programujaca:		; wypenianie bufora w pamici SRAM
	ldi_wsk Y, Bufor		;   inicjalizuj wskanik dla bufora w pamici SRAM
	ldi_wsk Z, PAGESIZE*2	;   inicjalizuj licznik bajtw w stronie

Wypelnij_bufor:
	rcall	Odbierz_transmisje;   odbierz bajt poprzez interfejs RS-232
	st	Y+, R16		;   zapisz go w buforze
	sbiw	ZL, 1			;   dekrementuj licznik pozostaych do odebrania bajtw
	brne	Wypelnij_bufor	;   zaptl iteracyjnie (a do odebrania penej strony)

					; przepisywanie wartoci ze SRAM do bufora strony flash
	ldi_wsk Y, Bufor		;   inicjalizuj wskanik dla bufora w pamici SRAM
	ldi_wsk Z, 0		;   inicjalizuj wskanik zapisu pamici flash
	ldi	R18, PAGESIZE	;   inicjalizuj licznik sw w stronie
Wypelnij_strone:
	ld	R0, Y+		;   zapisz w R0 LSB sowa z bufora SRAM
	ld	R1, Y+		;   zapisz w R1 MSB sowa z bufora SRAM

	ldi	R16, 1<<SPMEN
	rcall	Wykonaj_spm		;   zapisz odczytane sowo w buforze strony pamici flash

	adiw	ZL, 2			;   zwiksz wskanik zapisu pamici flash o 2
	dec	R18			;   zmniejsz licznik pozostaych do przepisania sw
	brne	Wypelnij_strone	;   zaptl iteracyjnie (a do wypenienia bufora strony)

	movw	ZL, STR_ZL		; zaaduj wskanik zapisu pamici flash adresem aktualnie
					;   zapisywanej strony  
	ldi	R16, (1<<PGWRT)|(1<<SPMEN)
	rcall	Wykonaj_spm		; zapisz stron w pamici flash

	ldi	R16, low(PAGESIZE<<1)
	add	STR_ZL, R16
	ldi	R16, high(PAGESIZE<<1)
	adc	STR_ZH, R16		; zwikszanie pomocniczego wskanika strony pamici
					;   flash tak, by wskazywa on na stron nastpn

	sbiw	XL, 1			; zmniejsz licznik pozostaych do zapisu stron
	brne	Petla_programujaca; zaptl iteracyjnie (a do zaprogramowania caej
					;   sekcji aplikacji)

	ldi	R16, (1<<RWWSRE)|(1<<SPMEN)
	rcall	Wykonaj_spm		; odblokuj moliwo odczytywania sekcji aplikacji


Koniec:				; ZAKOCZENIE DZIAANIA PROGRAMU BOOT LOADER-A
	cbi	K_LEDY, LED_ACT
	cbi	O_LEDY, LED_ACT	;   zapal wskanik aktywnoci boot loader-a

	ldi	R16, 'O'
	out	UDR, R16
	ldi	R16, 'K'
	out	UDR, R16		;   przelij wiadomo "OK" poprzez RS-232

	jmp	0			;   przejd do wykonywania programu aplikacji


Weryfikacja:			; WERYFIKACJA ZAWARTOCI SEKCJI APLIKACJI
	sbi	K_LEDY, LED_ACT
	sbi	O_LEDY, LED_ACT	; wcz wskanik aktywnoci boot loader-a

	ldi_wsk X, LARGEBOOTSTART<<1	; inicjalizuj licznik bajtw w segmencie aplikacji
	ldi_wsk Z, 0		; inicjalizuj wskanik odczytu pamici flash
Weryfikuj_kolejny:
	rcall	Odbierz_transmisje; odbierz bajt poprzez interfejs RS-232 (do R16)
	lpm	R17, Z+		; odczytaj bajt z pamici flash (do R17)

	cpse	R16, R17		; jeli warto odebrana jest rna od odczytanej, to..
	rjmp	Blad			;   zasygnalizuj bd

	cp	ZL, XL
	cpc	ZH, XH
	brne	Weryfikuj_kolejny	; zaptl iteracyjne (a do zweryfikowania caej
					;   sekcji aplikacji)
	rjmp	Koniec		; zakocz dziaanie boot loader-a
	

Odbierz_transmisje:		; ODBIR TRANSMISJI RS-232
	in	R16, UCSRA
	sbrs	R16, RXC		; oczekiwanie na odbir pojedynczego bajtu
	rjmp	Odbierz_transmisje

	andi	R16, (1<<FE)|(1<<DOR)|(1<<PE)
	brne	Blad			; jeli wystpi bd, to go zasygnalizuj
	in	R16, UDR		; zapisz odebrany bajt w rejestrze R16
	ret


Blad:					; SYGNALIZACJA BDU BOOT LOADER-A
	sbi	K_LEDY, LED_ERR
	sbi	O_LEDY, LED_ERR	; zapal wskanik bdu boot loader-a

	ldi	R16, 'E'
	out	UDR, R16
	ldi	R16, 'R'
	out	UDR, R16		; przelij wiadomo "ER"
Blad_stoj:
	rjmp	Blad_stoj		; zawie prac mikrokontrolera


Wykonaj_spm:			; WYKONYWANIE POLECENIA SPM
	in	R17, SPMCR
	sbrc	R17, SPMEN		; jeli operacja spm moe zosta wykonana, to..
	rjmp	Wykonaj_spm

	out	SPMCR, R16		;   przelij warto sterujc (R16) do rejestru SPMCR
	spm				;   i wykonaj operacj spm
	ret
