; PODPROGRAMY OBSUGI WYWIETLACZA LCD W TRYBIE 4-BITOWYM
; Rafa Baranowski (C) 2004
;
; Etykiety procedur:
; 	Ini_LCD		- inicjalizacja wywietlacza
;	Instrukcja_LCD	- przesyanie instrukcji do wywietlacza
;	Zapisz_LCD		- przesyanie danych do wywietlacza
;	Zapisz_CGRAM	- kopiowanie tablicy do pamici CGRAM
;	Czysc_LCD		- czyszczenie wywietlacza
;	Wroc_LCD		- powrt kursora do pozycji pocztkowej
;	Spacja_LCD		- generacja odstpu
;	Pokaz_bajt_LCD	- wywietlanie wartoci bajtu
;	Pokaz_ciag_LCD	- wywietlanie staego acucha znakw
;
; Wymagane stae:
;	K_LCD		- rejestr kierunku portu wyswietlacza
;	O_LCD		- rejestr wyjciowy portu wyswietlacza
;	LCD_LCD_RS	- nr linii dla sygnau LCD_RS
;	LCD_LCD_EN	- nr linii dla sygnau LCD_EN
;	LCD_LCD_DB4	- nr linii dla sygnau LCD_DB4
;	LCD_LCD_DB5	- nr linii dla sygnau LCD_DB5
;	LCD_LCD_DB6	- nr linii dla sygnau LCD_DB6
;	LCD_LCD_DB7	- nr linii dla sygnau LCD_DB7
;
; Wymagane podprogramy dodatkowe:
;	Czekaj_us	- z pliku czekaj_us.inc
;	Czekaj_ms	- z pliku czekaj_ms.inc


;****************************************************************
;		PROCEDURA INICJALIZUJCA WYWIETLACZ LCD
;
Ini_LCD:
	push	R16
	push	R17

	sbi	K_LCD, LCD_RS	; porty wywietlacza pracuj jako wyjcia
	sbi	K_LCD, LCD_EN
	sbi	K_LCD, LCD_DB4
	sbi	K_LCD, LCD_DB5
	sbi	K_LCD, LCD_DB6
	sbi	K_LCD, LCD_DB7
		
	cbi	O_LCD, LCD_RS	; przesyane wartoci bd instrukcjami
	cbi	O_LCD, LCD_EN

	ldi	R16, 20
	ldi	R17, 1
	rcall	Czekaj_ms		; oczekuwanie przez 20 ms na wypadek, gdyby
					;   zasilanie dopiero zostao wczone

	sbi	O_LCD, LCD_EN
	sbi	O_LCD, LCD_DB4
	sbi	O_LCD, LCD_DB5
	cbi	O_LCD, LCD_DB6
	cbi	O_LCD, LCD_DB7
	nop
	nop
	cbi	O_LCD, LCD_EN	; ustawienie DB7..4=0011 i potwierdzenie
					;   zboczem opadajcym na linii EN
	ldi	R16, 10
	ldi	R17, 1
	rcall	Czekaj_ms		; oczekiwanie przez 10 ms

	ldi	R17, 2
Ini_LCD_1:
	sbi	O_LCD, LCD_EN
	ldi	R16, 1
	rcall	Czekaj_us
	cbi	O_LCD, LCD_EN	
	ldi	R16, 20
	rcall	Czekaj_us
	dec	R17
	brne	Ini_LCD_1		; dwukrotne powtrzenie transmisji tej samej
					;   wartoci (DB7..4=0011) z opnieniem 200 us
					;   (programowa inicjalizacja wywietlacza)

	sbi	O_LCD, LCD_EN
	cbi	O_LCD, LCD_DB4
	ldi	R16, 1
	rcall	Czekaj_us
	cbi	O_LCD, LCD_EN	
	ldi	R16, 10
	rcall	Czekaj_us		; transmisja wartoci DB7..4=0010
					;   (przejcie w tryb interfejsu 4-bitowego)

	ldi	R16, (1<<LCD_USTAW)|(1<<LCD_U_4BIT)|(1<<LCD_U_2LIN)|(1<<LCD_U_5x8)
	rcall	Instrukcja_LCD	; interfejs 4-bitowy, 2 linie, 5x8 pkt

	ldi	R16, (1<<LCD_TRYB)|(1<<LCD_T_INC)
	rcall	Instrukcja_LCD	; tryb zwikszania adresu przy zapisie

	ldi	R16, (1<<LCD_STER)|(1<<LCD_SW_ON)|(1<<LCD_SK_ON)|(1<<LCD_SM_ON)
	rcall	Instrukcja_LCD	; wywietlanie wczone, kursor aktywny i miga

	sbi	O_LCD, LCD_RS	; przesyane wartoci bd danymi

	rcall	Czysc_LCD

	pop	R17
	pop	R16
	ret


;****************************************************************
;	PROCEDURA PRZESYAJCA INSTRUKCJ DO WYWIETLACZA
;
; Parametry:
;	R16	- warto instrukcji
;
; Przykad:
;				; przesuwanie wywietlacza w lewo
;	ldi	R16, (1<<LCD_PRZES)|(1<<LCD_P_WYS)|(1<<LCD_LEWO)
;	rcall	Instrukcja_LCD
;

				; etykiety symboliczne bitw uywanych w
				;   wartociach instrukcji
.EQU	LCD_CZYSC	= 0	; czyszczenie wywietlacza i powrt kursora
.EQU	LCD_WROC	= 1	; powrt kursora do pozycji pocztkowej

.EQU	LCD_TRYB	= 2	; zmiana trybu wywietlania
.EQU	LCD_T_INC	= 1	;   adres inkrementowany przy kadym zapisie
.EQU	LCD_T_DEC	= 2	;   adres dekrementowany przy kadym zapisie (domylnie)
.EQU	LCD_T_PRZES	= 0	;   wywietlacz przesuwany przy kadym zapisie

.EQU	LCD_STER	= 3	; sterowanie wywietlaczem
.EQU	LCD_SW_ON	= 2	;   wywietlanie wczone
.EQU	LCD_SW_OFF	= 3	;   wywietlanie wyczone (domylnie)
.EQU	LCD_SK_ON	= 1	;   kursor wczony
.EQU	LCD_SK_OFF	= 3	;   kursor wyczony (domylnie)
.EQU	LCD_SM_ON	= 0	;   miganie kursora wczone
.EQU	LCD_SM_OFF	= 3	;   miganie kursora wyczone (domylnie)

.EQU	LCD_PRZES	= 4	; przesuwanie
.EQU	LCD_P_KUR	= 4	;   przesuwanie kursora (domylnie)
.EQU	LCD_P_WYS	= 3	;   przesuwanie caej zawartoci wywietlacza
.EQU	LCD_LEWO	= 4	;   przesuwanie w lewo (domylnie)
.EQU	LCD_PRAWO	= 2	;   przesuwanie w prawo

.EQU	LCD_USTAW	= 5	; ustawienia interfejsu i wywietlacza
.EQU	LCD_U_8BIT	= 4	;   interfejs 8-bitowy
.EQU	LCD_U_4BIT	= 5	;   interfejs 4-bitowy (domylnie)
.EQU	LCD_U_2LIN	= 3	;   wywietlacz 2-liniowy
.EQU	LCD_U_1LIN	= 5	;   wywietlacz 1-liniowy (domylnie)
.EQU	LCD_U_5x10	= 2	;   znaki o rozmiarach 5x10 punktw
.EQU	LCD_U_5x8	= 5	;   znaki o rozmiarach 5x8 punktw (domylnie)

.EQU	LCD_CGADR	= 6	; zmiana adresu pamici generatora znakw (CGRAM)
.EQU	LCD_DDADR	= 7	; zmiana adresu pamici danych wywietlanych (DDRAM)

Instrukcja_LCD:
	cbi	O_LCD, LCD_RS	; przesyane wartoci bd instrukcjami
	rcall	Zapisz_LCD		; przelij instrukcj zawart w R16
	sbi	O_LCD, LCD_RS	; przesyane wartoci bd danymi

	ret


;***************************************************************
;			PROCEDURA PRZESYLAJCA DANE
;
; Parametry:
;	R16	- dane do wysania (np. kod wywietlanego znaku)
;
Zapisz_LCD:
	push	R16

	sbi	O_LCD, LCD_EN	; linia EN w stanie wysokim
	cbi	O_LCD, LCD_DB4
	cbi	O_LCD, LCD_DB5
	cbi	O_LCD, LCD_DB6
	cbi	O_LCD, LCD_DB7	; linie DB7..4 w stanie niskim
	sbrc	R16, 7
	sbi	O_LCD, LCD_DB7 
	sbrc	R16, 6
	sbi	O_LCD, LCD_DB6 
	sbrc	R16, 5
	sbi	O_LCD, LCD_DB5
	sbrc	R16, 4
	sbi	O_LCD, LCD_DB4	; ustawianie tylko tych linii DB7..4,
					;   ktrym odpowiadaj wysokie wartoci
					;   bitw 7..6 z rejestru R16
	cbi	O_LCD, LCD_EN	; linia EN w stanie niskim
					;   (przesanie starszego pbajtu)
	ldi	R16, 5
	rcall	Czekaj_us		; opnienie o 50 us

	pop	R16
	push	R16			; przywrcenie pocztkowej zawartoci R16

	sbi	O_LCD, LCD_EN
	cbi	O_LCD, LCD_DB4
	cbi	O_LCD, LCD_DB5
	cbi	O_LCD, LCD_DB6
	cbi	O_LCD, LCD_DB7
	sbrc	R16, 3
	sbi	O_LCD, LCD_DB7 
	sbrc	R16, 2
	sbi	O_LCD, LCD_DB6
	sbrc	R16, 1
	sbi	O_LCD, LCD_DB5
	sbrc	R16, 0
	sbi	O_LCD, LCD_DB4	; ustawianie tylko tych linii DB7..4,
					;   ktrym odpowiadaj wysokie wartoci
					;   bitw 3..0 z rejestru R16
	cbi	O_LCD, LCD_EN	; linia EN w stanie niskim
					;   (przesanie starszego pbajtu)
		
	ldi	R16, 5
	rcall	Czekaj_us		; opnienie o 50 us

	pop	R16
	ret


;****************************************************************
;	PROCEDURA KOPIUJCA TABLIC STAYCH DO PAMICI CGRAM
;
; Parametry:
;	Z	- adres pierwszej komrki tablicy w pamici programu
;
Zapisz_CGRAM:
	push	R16
	push	R17

	ldi	R16, 0|(1<<LCD_CGADR)
	rcall	Instrukcja_LCD	; wejcie w tryb adresowania pamici
					;   CGRAM i wymuszenie adresu zerowego

	ldi	R17, 64		; inicjalizacja licznika ptli (wyk. 64x)
Zapisz_CGRAM_1:
	lpm	R16, Z+
	rcall	Zapisz_LCD		; zapisywanie w pamici CGRAM kolejnych
					;   wartoci tablicy pamici programu,
					;   wskazywanej przez rejestr indeksowy Z
	dec	R17
	brne	Zapisz_CGRAM_1	; koniec ptli

	ldi	R16, 0|(1<<LCD_DDADR)
	rcall	Instrukcja_LCD	; wejcie w tryb adresowania pamici DDRAM
					;   (ustawienie kursora na pozycji (1,1))

	pop	R17
	pop	R16
	ret

			; ksztaty polskich znakw
ZNAK_A:	.db 0b00000000,0b00000000,0b00001110,0b00000001,0b00001111,0b00010001,0b00001111,0b00000011
ZNAK_C:	.db 0b00000010,0b00000100,0b00001110,0b00010000,0b00010000,0b00010001,0b00001110,0b00000000
ZNAK_E:	.db 0b00000000,0b00000000,0b00001110,0b00010001,0b00011111,0b00010000,0b00001110,0b00000011
ZNAK_L:	.db 0b00001100,0b00000100,0b00000110,0b00001100,0b00000100,0b00000100,0b00001110,0b00000000
ZNAK_N:	.db 0b00000010,0b00000100,0b00010110,0b00011001,0b00010001,0b00010001,0b00010001,0b00000000
ZNAK_O:	.db 0b00000010,0b00000100,0b00001110,0b00010001,0b00010001,0b00010001,0b00001110,0b00000000
ZNAK_S:	.db 0b00000010,0b00000100,0b00001110,0b00010000,0b00001110,0b00000001,0b00011110,0b00000000
ZNAK_Z:	.db 0b00000100,0b00000000,0b00011111,0b00000010,0b00000100,0b00001000,0b00011111,0b00000000

			; etykiety symboliczne polskich znakw po zaadowaniu
			;   pamici CGRAM wartociami z powyszej tablicy
.EQU POL_A	= 0	;   ""
.EQU POL_C	= 1	;   ""
.EQU POL_E	= 2	;   ""
.EQU POL_L	= 3	;   ""
.EQU POL_N	= 4	;   ""
.EQU POL_O	= 5	;   ""
.EQU POL_S	= 6	;   ""
.EQU POL_Z	= 7	;   ""


;***************************************************************
;		PROCEDURA CZYSZCZCA WYWIETLACZ LCD
;
Czysc_LCD:
	push	R16

	ldi	R16, (1<<LCD_CZYSC)
	rcall	Instrukcja_LCD	; czyszczenie wywietlacza i
					;   powrt kursora do pozycji
					;   pocztkowej
	ldi	R16, 200
	rcall	Czekaj_us		; opnienie o 2 ms

	pop	R16
	ret


;***************************************************************
;			PROCEDURA POWROTU KURSORA
;
Wroc_LCD:
	push	R16

	ldi	R16, (1<<LCD_WROC)
	rcall	Instrukcja_LCD	;   powrt kursora do pozycji
					;   pocztkowej
	ldi	R16, 200
	rcall	Czekaj_us		; opnienie o 2 ms

	pop	R16
	ret


;***************************************************************
;			PROCEDURA WPROWADZAJCA ODSTP
;
; Parametry:
;	R16	- liczba spacji do wywietlenia
;
Spacja_LCD:
	push	R16
	push	R17

	mov	R17, R16	; kopiuj liczb wymaganych spacji do R17
				;   (inicjalizacja licznika ptli)
	ldi	R16, ' '	; R16 zawiera kod ASCII pustego pola (spacji)
Spacja_LCD_1:
	rcall	Zapisz_LCD	; wywietlenie pustego pola (znaku z R16)
	dec	R17		; zmniejszenie licznika ptli
	brne	Spacja_LCD_1; iteracyjne zaptlenie

	pop	R17
	pop	R16
	ret


;***************************************************************
;		PROCEDURA WYWIETLAJCA WARTO BAJTU
;
; Parametry:
;	R16	- warto do wywietlenia
;
Pokaz_bajt_LCD:
	push	R16
	push	R17

	mov	R17, R16		; kopiuj warto wywietlan do R17
	clt				; zeruj znacznik T, ktry zostanie ustawiony,
					;   gdy natrafimy na pierwsz niezerow cyfr

	clr	R16			; ptla obliczajca cyfr setek
Pok_b_LCD_1:
	subi	R17, 100
	brlo	Pok_b_LCD_2
	inc	R16
	rjmp	Pok_b_LCD_1
Pok_b_LCD_2:
	subi	R17, -100
	tst	R16
	breq	Pok_b_LCD_3		; jeli liczba setek jest rna od 0, to..
	set				;   ustaw znacznik T (dalsze cyfry musz by
					;     wywietlane, nawet jeli s zerowe)
	ori	R16, 0b00110000	;   przekonwertuj j na kod ASCII (+48)
	rcall	Zapisz_LCD		;   i wywietl na LCD

	clr	R16			; ptla obliczajca cyfr dziesitek
Pok_b_LCD_3:
	subi	R17, 10
	brlo	Pok_b_LCD_4
	inc	R16
	rjmp	Pok_b_LCD_3
Pok_b_LCD_4:
	subi	R17, -10
	tst	R16
	brne	Pok_b_LCD_5		; jeli liczba dziesitek jest rwna 0, to..
	brtc	Pok_b_LCD_6		;   jeli znacznik T nie by ustawiony, to..
Pok_b_LCD_5:
	set				;     ustaw znacznik T (dalsze cyfry musz by
					;     wywietlane, nawet jeli s zerowe)
	ori	R16, 0b00110000	;     przekonwertuj j na kod ASCII (+48)
	rcall	Zapisz_LCD		;     i wywietl na LCD

Pok_b_LCD_6:
	mov	R16, R17		; pozostao jest cyfr jednostek
	ori	R16, 0b00110000	;   przekonwertuj j na kod ASCII (+48)
	rcall	Zapisz_LCD		;   i wywietl na LCD

	pop	R17
	pop	R16
	ret


;***************************************************************
;	PROCEDURA WYWIETLAJCA STAY ACUCH ZNAKW
;
; Parametry:
;	Z	- adres pierwszego znaku acucha w pamici programu

.EQU ZNAK_KONCA	= 0 	; etykieta symboliczna znaku koca acucha
				; (przyjmujemy 0, tzw. "null terminated string")

Pokaz_ciag_LCD:
	push	R16
	push	R17

Pok_c_LCD_1:		; ptla wywietlajca kolejne znaki ancucha
	lpm	R16, Z+	; odczytaj znak z pamici programu, spod adresu
				;   wskazywanego przez rejestr indeksowy Z
	cpi	R16, ZNAK_KONCA
	breq	Pok_c_LCD_2	; jeli znak jest znakiem koca, to zakocz,
	rcall	Zapisz_LCD	;   w przeciwnym razie wywietl go na LCD
	rjmp	Pok_c_LCD_1	;   i przejd do odczytywania kolejnego znaku

Pok_c_LCD_2:
	pop	R17
	pop	R16
	ret
