; PODPROGRAMY OBSUGI NADAJNIKA I ODBIORNIKA PODCZERWIENI (KOD RC5)
; Rafa Baranowski (C) 2004
;
; Etykiety procedur:
; 	Ini_IR	- inicjalizacja nadajnika i odbiornika podczerwieni
;	Odbior_IR	- odbir wiadomoci RC5 (obsuga INT1)
;	Nadaj_IR	- nadawanie wiadomoci RC5
;
; Wykorzystywane zasoby:
;	- licznik 2. (8-bitowy)
;	- rdo przerwa zewntrznych INT1
;
; Wymagane zmienne:
;	IR_ADRES	- zmienna bajtowa w pamici SRAM
;	IR_ROZKAZ	- zmienna bajtowa w pamici SRAM
;	IR_ZNACZ	- zmienna bajtowa w pamici SRAM
;
; Wymagane stae:
;	OSC_FREQ	- czstotliwo rezonatora kwarcowego [Hz]
;	K_IR_OUT	- rejestr kierunku portu nadajnika
;	O_IR_OUT	- rejestr wyjciowy portu nadajnika
;	IR_OUT	- nr linii dla sygnau nadawanego
;	K_IR_IN	- rejestr kierunku portu odbiornika 
;	O_IR_IN	- rejestr wyjciowy portu odbiornika
;	I_IR_IN	- rejestr wejciowy portu odbiornika
;	IR_IN		- nr linii dla sygnau odbieranego
;


;****************************************************************
;	PROCEDURA INICJALIZUJCA NADAJNIK I ODBIORNIK PODCZERWIENI
;
Ini_IR:
	push	R16

	sbi	K_IR_OUT, IR_OUT	; linia IR_OUT pracuje jako wyjcie
	cbi	O_IR_OUT, IR_OUT	; i znajduje si w stanie niskim
	
	cbi	K_IR_IN, IR_IN	; linia IR_IN pracuje jako wejcie
	sbi	O_IR_IN, IR_IN	; z aktywnym podciganiem

	in	R16, MCUCR
	sbr	R16, 1<<ISC11
	cbr	R16, 1<<ISC10
	out	MCUCR, R16		; przerwanie INT1 wywoywane zboczem opadajcym

	ldi	R16, 1<<INTF1
	out	GIFR, R16		; zeruj znacznik dania przerwania INT1

	in	R16, GICR
	sbr	R16, 1<<INT1
	out	GICR, R16		; wcz przerwanie INT1

	clr	R16
	sts	IR_ADRES, R16	; zeruj wykorzystywane zmienne w pamici SRAM
	sts	IR_ROZKAZ, R16
	sts	IR_ZNACZ, R16

	in	R16, TIMSK
	cbr	R16, (1<<OCIE2)|(1<<TOIE2)
	out	TIMSK, R16		; wycz obsug przerwa dla licznika 2.

	pop	R16
	ret


;****************************************************************
;	PROCEDURA ODBIERAJCA WIADOMO W KODZIE RC5 (obsuga INT1)
;
; Wymagane stae:
;	OSC_FREQ	- czstotliwo rezonatora kwarcowego [Hz]
;
; Wartoci wyjciowe:
;	IR_ADRES 	- zmienna adresu odebranej wiadomoci RC5 (SRAM)
;	IR_ROZKAZ	- zmienna rozkazu odebranej wiadomoci RC5 (SRAM)
;	IR_ZNACZ	- zmienna znacznikw wiadomoci RC5 (SRAM)
;			    (jeli odebrano now wiadomo, bit IR_NOWY
;			     tej zmiennej przyjmuje stan wysoki)
;

.EQU IR_NOWY = 0	; nr znacznika nowej wiadomoci w zmiennej IR_ZNACZ

Odbior_IR:
	push	R16
	push	R17
	push	R18
	push	R19
	push	XL
	push	XH
	in	R16, SREG
	push	R16

	ldi	R16, (1<<WGM21)|(1<<CS22)|(1<<CS20)
	out	TCCR2, R16		; licznik 2. pracuje w trybie CTC,
					;   zliczanie z czstotliwoci f=fosc/128
	ldi	R16, INT(0.00135*OSC_FREQ/128)
	out	OCR2, R16		; przepenianie z okresem ok. Tp=1,35 ms
					;   (3/4 czasu przesyu 1 bitu w kodzie RC5)
	ldi	R16, (1<<OCF2)
	out	TIFR, R16		; zerowanie znacznika zgodnego porwnania

	rcall Odb_IR_czekaj_na_OCF	; oczekiwanie na drugi bit startu

	sbis	I_IR_IN, IR_IN	; jeli bit startu jest niepoprawny (=0), to..
	rjmp	Odb_IR_koniec	;   zakocz procedur odbioru
	
	rcall	Odb_IR_czekaj_na_zmiane	; oczekuj na zmian sygnau odbieranego
	rcall Odb_IR_czekaj_na_OCF	; oczekuj na bit T (jest on ignorowany)

	clr	R18			; inicjalizacja rejestru adresu RC5 (5 bitw)
	ldi	R17, 5		; inicjalizacja licznika ptli (x5)
Odb_IR_adres:
	rcall	Odb_IR_czekaj_na_zmiane	; czekaj na zmian sygnau odbieranego
	rcall Odb_IR_czekaj_na_OCF	; oczekuj na kolejny bit adresu

	clc
	sbic	I_IR_IN, IR_IN	; kopiuj odebrany bit do znacznika C ze SREG
	sec
	rol	R18			; wsu odebrany bit (C) do rejestru adresu
	dec	R17
	brne	Odb_IR_adres	; zaptlenie iteracyjne

	clr	R19			; inicjalizacja rejestru rozkazu RC5 (6 bitw)
	ldi	R17, 6		; inicjalizacja licznika ptli (x6)
Odb_IR_rozkaz:
	rcall	Odb_IR_czekaj_na_zmiane	; czekaj na zmian sygnau odbieranego
	rcall Odb_IR_czekaj_na_OCF	; oczekuj na kolejny bit rozkazu

	clc
	sbic	I_IR_IN, IR_IN	; kopiuj odebrany bit do znacznika C ze SREG
	sec	
	rol	R19			; wsu odebrany bit (C) do rejestru rozkazu
	dec	R17
	brne	Odb_IR_rozkaz	; zaptlenie iteracyjne

	sts	IR_ADRES, R18	; zapisz odebrany adres RC5 w zmiennej w SRAM
	sts	IR_ROZKAZ, R19	; zapisz odebrany rozkaz RC5 w zmiennej w SRAM

	lds	R16, IR_ZNACZ	; odczytaj zmienn znacznikw wiadomoci RC5
	sbr	R16, 1<<IR_NOWY	; i ustaw w niej bit informujcy o odebraniu
	sts	IR_ZNACZ, R16	;   nowej wiadomoci (IR_NOWY)

Odb_IR_koniec:
	ldi	R16, (1<<INTF1)
	out	GIFR, R16		; zeruj znacznik dania przerwania INT1

	pop	R16
	out	SREG, R16
	pop	XH
	pop	XL
	pop	R19
	pop	R18
	pop	R17
	pop	R16
	reti

Odb_IR_czekaj_na_zmiane:	; pomocniczy podprogram oczekujcy na zmian
	sbic	I_IR_IN, IR_IN	;   stanu sygnau odbieranego
	rjmp	Odb_IR_cz_na_L
Odb_IR_cz_na_H:
	rcall	Odb_IR_spr_czas	;   sprawd, czy nie przekroczono czasu
	sbis	I_IR_IN, IR_IN
	rjmp	Odb_IR_cz_na_H	;   oczekiwanie na stan H, jeli by L
	ret
Odb_IR_cz_na_L:
	rcall	Odb_IR_spr_czas	;   sprawd, czy nie przekroczono czasu
	sbic	I_IR_IN, IR_IN
	rjmp	Odb_IR_cz_na_L	;   oczekiwanie na stan L, jeli by H
	ret

Odb_IR_spr_czas:			; pomocniczy podprogram sprawdzajcy, czy
	in	R16, TIFR		;   czas oczekiwania na zmian sygnau
	sbrs	R16, OCF2		;   odbieranego nie przekroczy 1,35 ms
	ret
					;   jeli przekroczy (wystpi bd), to..
	in	XH, SPH
	in	XL, SPL
	adiw	XH:XL, 4		;     zwiksz wskanik stosu o 4
	out	SPH, XH		;       (rwnowaenie stosu)
	out	SPL, XL
	rjmp	Odb_IR_koniec	;     i zakocz procedur odbierajc

Odb_IR_czekaj_na_OCF:		; pomocniczy podprogram wprowadzajcy
					;   okoo 1.35ms opnienia
	ldi	R16, 0
	out	TCNT2, R16		;   zeruj zawartoci licznika 2.

Odb_IR_cz_na_OCF_1:
	in	R16, TIFR		
	sbrs	R16, OCF2		;   czekaj na przepenienie licznika 2.
	rjmp	Odb_IR_cz_na_OCF_1

	ldi	R16, (1<<OCF2)
	out	TIFR, R16		;   zeruj znacznik zgodnego porwnania
	ret


;****************************************************************
;	PROCEDURA NADAJCA WIADOMO W KODZIE RC5
;
; Wymagane stae:
;	OSC_FREQ	- czstotliwo rezonatora kwarcowego [Hz]
;
; Parametry:
;	R16	- rejestr adresu nadawanej wiadomoci RC5 (zakres 0..31)
;	R17	- rejestr rozkazu nadawanej wiadomoci RC5 (zakres 0..63)
;
; Wykorzystuje:
;	IR_ZNACZ	- zmienna znacznikw wiadomoci RC5 (SRAM)
;
; Uwagi:
;	- podprogram automatycznie wcza zezwolenie na obsug przerwa
;

.EQU IR_T = 1	; nr znacznika T w zmiennej IR_ZNACZ

Nadaj_IR:
	push	R18
	push	R19

	push	R17			; umie rozkaz wiadomoci do nadania na stosie
	push	R16			; umie adres wiadomoci do nadania na stosie

	cli				; na czas nadawania wycz obsug przerwa

	ldi	R16, INT(OSC_FREQ/(2*36000))
	out	OCR2, R16		; czstotliwo przepenie licznika 2. rwna
					;   okoo 2x36 kHz (cz. generowana na OC2
					;   jest zatem rwna 36 kHz)
	rcall	Nad_IR_H		; nadaj 1. bit startu (bit H)
	rcall	Nad_IR_H		; nadaj 2. bit startu (bit H)

	lds	R16, IR_ZNACZ	; odczytaj zmienn stanu wiadomoci RC5,
	ldi	R17, 1<<IR_T
	eor	R16, R17		; zaneguj w niej bit T
	sts	IR_ZNACZ, R16	; i ponownie zapisz w pamici SRAM
	sbrs	R16, IR_T		; jeli warto bitu T bya wysoka, to..
	rjmp	Nad_IR_1
	rcall	Nad_IR_H		;   nadaj bit o wartoci H
	rjmp	Nad_IR_2
Nad_IR_1:
	rcall	Nad_IR_L		;   w przeciwnym razie - nadaj L
Nad_IR_2:

	pop	R19			; cignij ze stosu adres do nadania
					;   i zapisz go w rejestrze R19
	ldi	R18, 1<<4		; inicjalizacja licznika ptli iteracyjnej (x5)
Nad_IR_adres:
	mov	R16, R19
	and	R16, R18
	breq	Nad_IR_adres_1	; w zalenoci od stanu kolejnych bitw
	rcall	Nad_IR_H		;   nadawanego adresu RC5, nadaj bit L lub H
	rjmp	Nad_IR_adres_2	
Nad_IR_adres_1:
	rcall	Nad_IR_L
Nad_IR_adres_2:
	lsr	R18
	brne	Nad_IR_adres	; zaptlenie iteracyjne

	pop	R19			; cignij ze stosu rozkaz do nadania
					;   i zapisz go w rejestrze R19
	ldi	R18, 1<<5		; inicjalizacja licznika ptli iteracyjnej (x6)
Nad_IR_rozkaz:
	mov	R16, R19
	and	R16, R18
	breq	Nad_IR_rozkaz_1	; w zalenoci od stanu kolejnych bitw
	rcall	Nad_IR_H		;   nadawanego rozkazu RC5, nadaj bit L lub H
	rjmp	Nad_IR_rozkaz_2	
Nad_IR_rozkaz_1:
	rcall	Nad_IR_L
Nad_IR_rozkaz_2:
	lsr	R18
	brne	Nad_IR_rozkaz	; zaptlenie iteracyjne

	rcall	Nad_IR_nieakt	; wycz nadajnik

	pop	R19
	pop	R18
	reti				; zakocz i przywr obsug przerwa (!)

Nad_IR_H:				; pomocniczy podprogram nadajcy bit H
	rcall	Nad_IR_nieakt	; wycz nadajnik
	rcall	Nad_IR_900us	; czekaj 900 us

	rcall	Nad_IR_akt		; wcz nadajnik
	rcall	Nad_IR_900us	; czekaj 900 us
	ret

Nad_IR_L:				; pomocniczy podprogram nadajcy bit L
	rcall	Nad_IR_akt		; wcz nadajnik
	rcall	Nad_IR_900us	; czekaj 900 us

	rcall	Nad_IR_nieakt	; wycz nadajnik
	rcall	Nad_IR_900us	; czekaj 900 us
	ret

Nad_IR_akt:				; pomocniczy podprogram wczajcy nadajnik
	ldi	R16, (1<<WGM21)|(1<<CS20)|(1<<COM20)
	out	TCCR2, R16		; licznik 2. pracuje w trybie CTC, f=fosc,
	ret				;   przy przepenianiu wyjcie OC2 zmienia stan
					;   na przeciwny

Nad_IR_nieakt:			; pomocniczy podprogram wyczajcy nadajnik
	ldi	R16, (1<<WGM21)|(1<<CS20)
	out	TCCR2, R16		; licznik 2. pracuje w trybie CTC, f=fosc,
	ret				;   bez wzgldu na dziaanie licznika, wyjcie 
					;   OC2 pozostaje nieaktywne (w stanie L)

Nad_IR_900us:			; pomocniczy podprogram oczekujcy ok. 900 us
	ldi	R17, 65		; inicjalizacja licznika ptli iteracyjnej (x65)
Nad_IR_900us_1:
	in	R16, TIFR
	sbrs	R16, OCF2		; oczekuj na przepenienie licznika 2.
	rjmp	Nad_IR_900us_1

	ldi	R16, (1<<OCF2)
	out	TIFR, R16		; zeruj znacznik zgodnego porwnania

	dec	R17
	brne	Nad_IR_900us_1	; zaptlenie iteracyjne

	ret
