; Woltomierz cyfrowy, obsuga przetwornika ADC
; Rafa Baranowski (C) 2004

.INCLUDE "m16def.inc"		; SEGMENT DEKLARACJI

.EQU SYS_FREQ	= 16		; czstotliwo pracy w MHz
					; stae klawiatury matrycowej
.EQU K_KLAW		= DDRC	;   rejestr kierunku portu
.EQU O_KLAW		= PORTC	;   rejestr wyjciowy portu
.EQU I_KLAW		= PINC	;   rejestr wejciowy portu
.EQU K_KL_WCISN	= DDRD	;   rejestr kierunku linii "kl. wcinity"
.EQU O_KL_WCISN	= PORTD	;   rejestr wyjciowy linii "kl. wcinity"
.EQU I_KL_WCISN	= PIND	;   rejestr wejciowy linii "kl. wcinity"
.EQU KL_WCISN	= 2		;   nr linii dla sygnau "kl. wcinity"
					; stae wywietlacza multipleksowanego LED
.EQU K_W7S_SEG	= DDRB	;   rejestr kierunku portu segmentw
.EQU O_W7S_SEG	= PORTB	;   rejestr wyjciowy portu segmentw
.EQU K_W7S_KOL	= DDRD	;   rejestr kierunku portu kolumn
.EQU O_W7S_KOL	= PORTD	;   rejestr wyjciowy portu kolumn
.EQU W7S_KOL0	= 7		;   nr linii dla kolumny 1. (cyfra najmodsza)
.EQU W7S_KOL1	= 6		;   nr linii dla kolumny 2.
.EQU W7S_KOL2	= 1		;   nr linii dla kolumny 3.
.EQU W7S_KOL3	= 0		;   nr linii dla kolumny 4. (cyfra najstarsza)


.DSEG					; SEGMENT DANYCH
.ORG	0x0060			; pocztek pamici SRAM
Klaw_inf:		.BYTE	1	; zmienna stanu klawiatury matrycowej
W7s_kolumna:	.BYTE 1	; zmienna aktywnej kolumny wywietlacza LED
W7s_cyfry:		.BYTE 4	; zmienne cyfr wywietlacza LED


.CSEG					; SEGMENT PROGRAMU
.ORG	0				; wektor zerowania
	jmp	Reset			; skok do procedury inicjalizujcej
.ORG	INT0addr			; wektor przerwania INT0
	jmp	Odczyt_klawiatury	; skok do procedury odczytujcej stan klawiatury
.ORG	OC0addr			; wektor przerwania OC0
	jmp	Multipleks_w7s	; skok do procedury obsugi wywietlacza multipl.
.ORG	ADCCaddr			; wektor przerwania ADCC
	reti				; procedura obsugi przerwania ADC jest pusta

.ORG	0x30				; przejcie do pierwszej komrki za wektorami przerwa


Reset:				; podprogram inicjalizujcy
	ldi	R17, high(RAMEND)
	ldi	R16, low(RAMEND)
	out	SPH, R17
	out	SPL, R16		; inicjalizacja wskanika stosu
					; inicjalizacja zawartoci pamici SRAM
	clr	R16
	ldi	R17, 6		;   R17 okrela liczb zajtych przez zmienne komrek
	ldi	XH, high(0x0060)	;     pamici SRAM (licznik poniszej ptli)
	ldi	XL, low(0x0060)	;   X wskazuje na pocztek pamici SRAM
Reset_1:				;   ptla zerujca zawarto wszystkich uywanych
	st	X+, R16		;     przez zmienne komrek pamici SRAM
	dec	R17
	brne	Reset_1

	rcall	Ini_klawiatury	; inicjalizacja klawiatury matrycowej
	rcall	Ini_w7s		; inicjalizacja wywietlacza LED

	ldi	R16, (1<<REFS1)|(1<<REFS0)|(1<<ADLAR)
	out	ADMUX, R16		; przetwornik ADC pracuje z wewntrznym rdem
					;   napicia odniesienia (2,56V), a wynik konwersji
					;   przedstawiany jest w postaci przesunitej w lewo
	ldi	R16, (1<<ADEN)|(1<<ADIE)|(1<<ADPS2)|(1<<ADPS1)|(1<<ADPS0)
	out	ADCSR, R16		; przetwornik ADC pracuje w trybie jednorazowego
					;   wyzwalania, przy taktowaniu f=fosc/128
					;   (125 kHz), przerwania dla ADC s aktywne
	in	R16, MCUCR
	cbr	R16, (1<<SM2)|(1<<SM1)
	sbr	R16, (1<<SM0)|(1<<SE)
	out	MCUCR, R16		; tryby upienia s dozwolone; trybem domylnym jest
					;   upienie z redukcj szumw dla ADC
	sei				; odblokowanie przerwa


Petla:				; ptla gwna programu
	lds	R16, Klaw_inf
	sbrc	R16, 7		; jeli nacinito dowolny klawisz,
	rjmp	P_zmien_kanal	;   to przejd do procedury zmiany kanau ADC

	ldi	R16, 100
	ldi	R17, 2
	rcall	Czekaj_ms		; w przeciwnym razie czekaj 200 ms
	rjmp	P_konwersja		; i przejd do przeprowadzania konwersji

P_zmien_kanal:			; jeli nacinito klawisz, to..
	lds	R16, Klaw_inf
	sbrc	R16, 7		;   oczekuj na jego puszczenie
	rjmp	P_zmien_kanal

	cpi	R16, 8		;   jeli nr klawisza jest wikszy od 7, to..
	brsh	P_konwersja		;     nie zmieniaj aktualnego kanau ADC i
					;     przejd do przeprowadzania konwersji
	ori	R16, (1<<REFS1)|(1<<REFS0)|(1<<ADLAR)
	out	ADMUX, R16		;   w przeciwnym razie przecz multiplekser ADC
					;     na kana ADC o numerze odpowiadajcym
					;     wcinitemu klawiszowi

P_konwersja:			; konwersja ADC
	in	R16, TIMSK
	cbr	R16, 1<<OCIE0
	out	TIMSK, R16		; wycz obsug przerwania OCIE0 (wyw. LED)

	sleep				; przejd w tryb upienia z automatycznym
					;   uruchomieniem jednorazowej konwersji
	in	R16, TIMSK
	sbr	R16, 1<<OCIE0
	out	TIMSK, R16		; przywr obsug przerwania OCIE0 (wyw. LED)

	in	R16, ADCH
	rcall	Pokaz_bajt_w7s	; poka wynik konwersji na wywietlaczu LED

	rjmp	Petla			; zaptl program


					; DOCZANE PODPROGRAMY
.INCLUDE "czekaj_ms.inc"	;   oczekiwanie milisekundowe
.INCLUDE "czekaj_us.inc"	;   oczekiwanie mikrosekundowe
.INCLUDE "klaw_mat.inc"		;   obsuga klawiatury matrycowej
.INCLUDE "w7s_mul.inc"		;   wywietlacz multipleksowany LED
